/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer;

import static org.lwjgl.opengl.GL11.GL_FLOAT;

import java.awt.Color;
import java.nio.Buffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;

import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.GL44;
import org.lwjgl.opengl.GL45;
import org.lwjgl.system.MemoryUtil;

import de.hft.stuttgart.citygml.viewer.datastructure.BufferConstants;
import de.hft.stuttgart.citygml.viewer.datastructure.Polygon;
import de.hft.stuttgart.citygml.viewer.datastructure.TesselatedPolygon;
import de.hft.stuttgart.citygml.viewer.datastructure.Triangle3d;
import de.hft.stuttgart.citygml.viewer.math.Vector3d;

public class PolygonViewInformation {
	
	private static final Vector3d AXIS = new Vector3d(19, 0.8, 1.5).normalize();


	private int vao;
	private IntBuffer buffers;
	private boolean draw = false;
	private int elements;
	
	private PolygonListener listener;

	public PolygonViewInformation(List<Polygon> polygons, PolygonListener listener) {
		this.listener = listener;
		vao = GL45.glCreateVertexArrays();
		GL45.glVertexArrayAttribBinding(vao, BufferConstants.ATTRIBUTE_POSITION, BufferConstants.BINDING_INDEX_0);
		GL45.glVertexArrayAttribBinding(vao, BufferConstants.ATTRIBUTE_COLOR, BufferConstants.BINDING_INDEX_0);

		GL45.glVertexArrayAttribFormat(vao, BufferConstants.ATTRIBUTE_POSITION, 3, GL_FLOAT, false, 0);
		GL45.glVertexArrayAttribFormat(vao, BufferConstants.ATTRIBUTE_COLOR, 3, GL_FLOAT, false, 3 * 4);

		GL45.glEnableVertexArrayAttrib(vao, BufferConstants.ATTRIBUTE_POSITION);
		GL45.glEnableVertexArrayAttrib(vao, BufferConstants.ATTRIBUTE_COLOR);

		buffers = MemoryUtil.memAllocInt(BufferConstants.SIZE);
		GL45.glCreateBuffers(buffers);
		GL45.glVertexArrayElementBuffer(vao, buffers.get(BufferConstants.ELEMENT));
		GL45.glVertexArrayVertexBuffer(vao, BufferConstants.BINDING_INDEX_0, buffers.get(BufferConstants.VERTEX), 0, (3 + 3) * 4);
		setupBuffers(polygons);
	}
	
	private void setupBuffers(List<Polygon> polys) {
		List<TesselatedPolygon> tessPolys = new ArrayList<>();
		for (Polygon poly : polys) {
			tessPolys.add(TesselatedPolygon.of(poly));
			listener.finishedPolygon();
		}
		
		int nrOfTriangles = 0;
		for (TesselatedPolygon tessPoly : tessPolys) {
			nrOfTriangles += tessPoly.getTriangles().size();
		}

		// 6 float for 3 vertices per triangle
		int moreSpace = nrOfTriangles * 3 * 3 * 2;
		FloatBuffer vertexData = MemoryUtil.memAllocFloat(moreSpace);
		IntBuffer elementData = MemoryUtil.memAllocInt(nrOfTriangles * 3);
		float[] rgba = new float[4];
		for (TesselatedPolygon tessPoly : tessPolys) {
			Color baseColor = tessPoly.getColor();
			double cos = tessPoly.getNormal().dot(AXIS);
			double acos = Math.acos(cos);
			// normalize to range [0.3, 0.9]
			acos = acos / Math.PI;
			acos = acos * 0.6 + 0.3;
			Color derivedColor = new Color((int) (baseColor.getRed() * acos), (int) (baseColor.getGreen() * acos),
					(int) (baseColor.getBlue() * acos));
			derivedColor.getComponents(rgba);
			for (Triangle3d tri : tessPoly.getTriangles()) {
				addPoint(vertexData, tri.getP1(), rgba[0], rgba[1], rgba[2]);
				addPoint(vertexData, tri.getP2(), rgba[0], rgba[1], rgba[2]);
				addPoint(vertexData, tri.getP3(), rgba[0], rgba[1], rgba[2]);
			}
			listener.finishedPolygon();
		}
		for (int i = 0; i < elementData.capacity(); i++) {
			elementData.put(i);
		}
		((Buffer) elementData).flip();
		((Buffer) vertexData).flip();

		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, buffers.get(BufferConstants.VERTEX));
		GL44.glBufferStorage(GL15.GL_ARRAY_BUFFER, vertexData, 0);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, 0);

		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, buffers.get(BufferConstants.ELEMENT));
		GL44.glBufferStorage(GL15.GL_ELEMENT_ARRAY_BUFFER, elementData, 0);
		GL15.glBindBuffer(GL15.GL_ELEMENT_ARRAY_BUFFER, 0);
		
		MemoryUtil.memFree(elementData);
		MemoryUtil.memFree(vertexData);

		draw = true;
		elements = elementData.capacity();
	}
	
	private static void addPoint(FloatBuffer lod1VertexData, Vector3d point, float red, float green, float blue) {
		lod1VertexData.put((float) point.getX());
		lod1VertexData.put((float) point.getY());
		lod1VertexData.put((float) point.getZ());
		lod1VertexData.put(red);
		lod1VertexData.put(green);
		lod1VertexData.put(blue);
	}
	
	public boolean shouldDraw() {
		return draw;
	}
	
	public void setShouldDraw(boolean draw) {
		this.draw = draw;
	}
	
	public int getVao() {
		return vao;
	}
	
	public void destroy() {
		MemoryUtil.memFree(buffers);
	}

	public void draw() {
		if (draw) {
			GL30.glBindVertexArray(vao);
			GL11.glDrawElements(GL11.GL_TRIANGLES, elements, GL11.GL_UNSIGNED_INT, 0);
		}		
	}
}
