/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer.parser;

import java.awt.Color;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.citygml4j.model.bridge.AbstractBridge;
import org.citygml4j.model.building.AbstractBuilding;
import org.citygml4j.model.building.BuildingInstallation;
import org.citygml4j.model.construction.AbstractFillingElement;
import org.citygml4j.model.construction.Door;
import org.citygml4j.model.construction.GroundSurface;
import org.citygml4j.model.construction.RoofSurface;
import org.citygml4j.model.construction.Window;
import org.citygml4j.model.core.AbstractThematicSurface;
import org.citygml4j.model.core.ImplicitGeometry;
import org.citygml4j.model.core.ImplicitGeometryProperty;
import org.citygml4j.model.deprecated.bridge.DeprecatedPropertiesOfAbstractBridge;
import org.citygml4j.model.deprecated.building.DeprecatedPropertiesOfAbstractBuilding;
import org.citygml4j.model.deprecated.building.DeprecatedPropertiesOfBuildingInstallation;
import org.citygml4j.model.deprecated.core.DeprecatedPropertiesOfAbstractThematicSurface;
import org.citygml4j.model.deprecated.transportation.DeprecatedPropertiesOfAbstractTransportationSpace;
import org.citygml4j.model.deprecated.transportation.TransportationComplex;
import org.citygml4j.model.deprecated.vegetation.DeprecatedPropertiesOfPlantCover;
import org.citygml4j.model.deprecated.vegetation.DeprecatedPropertiesOfSolitaryVegetationObject;
import org.citygml4j.model.deprecated.waterbody.DeprecatedPropertiesOfWaterBody;
import org.citygml4j.model.landuse.LandUse;
import org.citygml4j.model.relief.TINRelief;
import org.citygml4j.model.transportation.AuxiliaryTrafficArea;
import org.citygml4j.model.transportation.TrafficArea;
import org.citygml4j.model.vegetation.PlantCover;
import org.citygml4j.model.vegetation.SolitaryVegetationObject;
import org.citygml4j.model.waterbody.WaterBody;
import org.citygml4j.visitor.ObjectWalker;
import org.locationtech.proj4j.BasicCoordinateTransform;
import org.locationtech.proj4j.ProjCoordinate;
import org.xmlobjects.gml.model.geometry.AbstractGeometry;
import org.xmlobjects.gml.model.geometry.GeometricPositionList;
import org.xmlobjects.gml.model.geometry.GeometryProperty;
import org.xmlobjects.gml.model.geometry.primitives.AbstractRing;
import org.xmlobjects.gml.model.geometry.primitives.AbstractRingProperty;
import org.xmlobjects.gml.model.geometry.primitives.LinearRing;
import org.xmlobjects.gml.model.geometry.primitives.Triangle;
import org.xmlobjects.gml.model.geometry.primitives.TriangulatedSurface;

import de.hft.stuttgart.citygml.viewer.datastructure.Polygon;
import de.hft.stuttgart.citygml.viewer.datastructure.Ring;
import de.hft.stuttgart.citygml.viewer.math.Vector3d;

public class FeatureMapper extends ObjectWalker {

	private Color groundColor = new Color(0.9411765f, 0.9019608f, 0.54901963f);
	private Color roofColor = Color.RED;
	private Color doorColor = Color.ORANGE;
	private Color windowColor = new Color(0.0f, 0.5019608f, 0.5019608f);
	private Color wallColor = Color.WHITE;
	private Color bridgeColor = new Color(1.0f, 0.49803922f, 0.3137255f);
	private Color landColor = new Color(0.64705884f, 0.16470589f, 0.16470589f);
	private Color transportationColor = new Color(1.0f, 1.0f, 0.0f);
	private Color vegetationColor = new Color(0.5647059f, 0.93333334f, 0.5647059f);
	private Color waterColor = new Color(0.5294118f, 0.80784315f, 0.98039216f);

	private ParserConfiguration config;
	private List<Polygon> lod1Polygons;
	private List<Polygon> lod2Polygons;
	private List<Polygon> lod3Polygons;
	private List<Polygon> lod4Polygons;

	private List<Polygon> currentPolygons = null;
	private Ring currentRing;
	private Color currentColor;

	private ProjCoordinate p1 = new ProjCoordinate();
	private ProjCoordinate p2 = new ProjCoordinate();

	public FeatureMapper(ParserConfiguration config) throws IOException {
		try (FileReader reader = new FileReader("color.properties")) {
			Properties props = new Properties();
			props.load(reader);
			Color parsedGroundColor = parseColor(props.getProperty("groundColor"));
			if (parsedGroundColor != null) {
				groundColor = parsedGroundColor;
			}
			Color parsedRoofColor = parseColor(props.getProperty("roofColor"));
			if (parsedRoofColor != null) {
				roofColor = parsedRoofColor;
			}
			Color parsedDoorColor = parseColor(props.getProperty("doorColor"));
			if (parsedDoorColor != null) {
				doorColor = parsedDoorColor;
			}
			Color parsedWindowColor = parseColor(props.getProperty("windowColor"));
			if (parsedWindowColor != null) {
				windowColor = parsedWindowColor;
			}
			Color parsedWallColor = parseColor(props.getProperty("wallColor"));
			if (parsedWallColor != null) {
				wallColor = parsedWallColor;
			}
			Color parsedBridgeColor = parseColor(props.getProperty("bridgeColor"));
			if (parsedBridgeColor != null) {
				bridgeColor = parsedBridgeColor;
			}
			Color parsedLandColor = parseColor(props.getProperty("landColor"));
			if (parsedLandColor != null) {
				landColor = parsedLandColor;
			}
			Color parsedTransportationColor = parseColor(props.getProperty("transportationColor"));
			if (parsedTransportationColor != null) {
				transportationColor = parsedTransportationColor;
			}
			Color parsedVegetationColor = parseColor(props.getProperty("vegetationColor"));
			if (parsedVegetationColor != null) {
				vegetationColor = parsedVegetationColor;
			}
			Color parsedWaterColor = parseColor(props.getProperty("waterColor"));
			if (parsedWaterColor != null) {
				waterColor = parsedWaterColor;
			}
		}
		this.config = config;
		lod1Polygons = new ArrayList<>();
		lod2Polygons = new ArrayList<>();
		lod3Polygons = new ArrayList<>();
		lod4Polygons = new ArrayList<>();
	}

	private Color parseColor(String colorString) {
		if (colorString == null) {
			return null;
		}
		try {
			String[] split = colorString.split(" ");
			float r = Float.parseFloat(split[0]);
			float g = Float.parseFloat(split[1]);
			float b = Float.parseFloat(split[2]);
			return new Color(r, g, b);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	@Override
	public void visit(org.xmlobjects.gml.model.geometry.primitives.Polygon gmlPoly) {
		if (currentPolygons == null) {
			return;
		}
		Polygon viewerPoly = new Polygon(currentColor);
		// parse rings
		Ring extRing = new Ring();
		viewerPoly.setExteriorRing(extRing);
		mapRing(gmlPoly.getExterior(), extRing);

		if (gmlPoly.getInterior() != null) {
			for (AbstractRingProperty arp : gmlPoly.getInterior()) {
				if (arp.getObject() != null) {
					Ring innerRing = new Ring();
					viewerPoly.addInteriorRing(innerRing);
					mapRing(arp, innerRing);
				}
			}
		}
		currentPolygons.add(viewerPoly);
	}
	
	@Override
	public void visit(Triangle triangle) {
		if (currentPolygons == null) {
			return;
		}
		// triangle is only a special case of a polygon
		// so use a polygon
		Polygon viewerPoly = new Polygon(currentColor);
		// parse rings
		Ring extRing = new Ring();
		viewerPoly.setExteriorRing(extRing);
		mapRing(triangle.getExterior(), extRing);
		currentPolygons.add(viewerPoly);
	}
	
	private void mapRing(AbstractRingProperty gmlRing, Ring viewerRing) {
		if (gmlRing == null || gmlRing.getObject() == null) {
			return;
		}
		AbstractRing ringGeometry = gmlRing.getObject();
		currentRing = viewerRing;
		// jump to LinearRing or Ring visit
		ringGeometry.accept(this);
	}

	@Override
	public void visit(LinearRing linearRing) {
		if (linearRing.getControlPoints() != null) {
			int dimension = getDimension(linearRing);
			parseControlPoints(linearRing.getControlPoints(), dimension);
		}
	}

	private int getDimension(LinearRing linearRing) {
		int dimension = 3;
		if (linearRing.getSrsDimension() != null) {
			dimension = linearRing.getSrsDimension();
		}
		return dimension;
	}

	private void parseControlPoints(GeometricPositionList controlPoints, int dimension) {
		List<Double> coords = controlPoints.toCoordinateList3D();
		switch (dimension) {
		case 1:
			for (int i = 0; i < coords.size(); i++) {
				createVertex(coords.get(i), 0, 0);
			}
			break;
		case 2:
			for (int i = 0; i < coords.size(); i = i + 2) {
				createVertex(coords.get(i + 0), coords.get(i + 1), 0);
			}
			break;
		case 3:
			for (int i = 0; i < coords.size(); i = i + 3) {
				createVertex(coords.get(i + 0), coords.get(i + 1), coords.get(i + 2));
			}
			break;
		default:
			throw new UnsupportedOperationException("Cannot parse Coordinates with dimension:" + dimension);
		}
	}

	private void createVertex(double x, double y, double z) {
		// transform into utm, if available
		BasicCoordinateTransform trans = config.getTargetTransform();
		if (trans != null) {
			p1.setValue(x, y);
			trans.transform(p1, p2);
			x = p2.x;
			y = p2.y;
		}
		Vector3d v = new Vector3d(x, y, z);
		currentRing.addPoint(v);
	}

	@Override
	public void visit(AbstractBuilding ab) {
		super.visit(ab);
		DeprecatedPropertiesOfAbstractBuilding deprecatedProperties = ab.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSurface(), lod1Polygons, wallColor);
		parseAbstractGeometry(ab.getLod2MultiSurface(), lod2Polygons, wallColor);
		parseAbstractGeometry(ab.getLod3MultiSurface(), lod3Polygons, wallColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, wallColor);
		parseAbstractGeometry(ab.getLod1Solid(), lod1Polygons, wallColor);
		parseAbstractGeometry(ab.getLod2Solid(), lod2Polygons, wallColor);
		parseAbstractGeometry(ab.getLod3Solid(), lod3Polygons, wallColor);
		parseAbstractGeometry(deprecatedProperties.getLod4Solid(), lod4Polygons, wallColor);
	}
	
	@Override
	public void visit(TINRelief tinRelief) {
		super.visit(tinRelief);
		if (tinRelief.getTin() == null || tinRelief.getTin().getObject() == null) {
			return;
		}
		TriangulatedSurface object = tinRelief.getTin().getObject();
		currentPolygons = lod1Polygons;
		currentColor = landColor;
		object.accept(this);
		currentPolygons = null;
		currentColor = null;
	}

	@Override
	public void visit(WaterBody wb) {
		super.visit(wb);
		DeprecatedPropertiesOfWaterBody deprecatedProperties = wb.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSurface(), lod1Polygons, waterColor);
		parseAbstractGeometry(wb.getLod1Solid(), lod1Polygons, waterColor);
		parseAbstractGeometry(wb.getLod2Solid(), lod2Polygons, waterColor);
		parseAbstractGeometry(wb.getLod3Solid(), lod3Polygons, waterColor);
		parseAbstractGeometry(deprecatedProperties.getLod4Solid(), lod4Polygons, waterColor);
	}

	@Override
	public void visit(PlantCover pc) {
		super.visit(pc);
		DeprecatedPropertiesOfPlantCover deprecatedProperties = pc.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSurface(), lod1Polygons, vegetationColor);
		parseAbstractGeometry(pc.getLod2MultiSurface(), lod2Polygons, vegetationColor);
		parseAbstractGeometry(pc.getLod3MultiSurface(), lod3Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSolid(), lod1Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod2MultiSolid(), lod2Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod3MultiSolid(), lod3Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSolid(), lod4Polygons, vegetationColor);
	}

	@Override
	public void visit(SolitaryVegetationObject svo) {
		super.visit(svo);
		DeprecatedPropertiesOfSolitaryVegetationObject deprecatedProperties = svo.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1Geometry(), lod1Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod2Geometry(), lod2Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod3Geometry(), lod3Polygons, vegetationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4Geometry(), lod4Polygons, vegetationColor);
		parseImplicitGeometry(svo.getLod1ImplicitRepresentation(), lod1Polygons, vegetationColor);
		parseImplicitGeometry(svo.getLod2ImplicitRepresentation(), lod2Polygons, vegetationColor);
		parseImplicitGeometry(svo.getLod3ImplicitRepresentation(), lod3Polygons, vegetationColor);
		parseImplicitGeometry(deprecatedProperties.getLod4ImplicitRepresentation(), lod4Polygons, vegetationColor);
	}

	@Override
	public void visit(AbstractThematicSurface abs) {
		super.visit(abs);
		Color color = getBuildingSurfaceColor(abs);
		DeprecatedPropertiesOfAbstractThematicSurface deprecatedProperties = abs.getDeprecatedProperties();
		parseAbstractGeometry(abs.getLod2MultiSurface(), lod2Polygons, color);
		parseAbstractGeometry(abs.getLod3MultiSurface(), lod3Polygons, color);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, color);
	}

	@Override
	public void visit(LandUse landUse) {
		super.visit(landUse);
		DeprecatedPropertiesOfAbstractThematicSurface deprecatedProperties = landUse.getDeprecatedProperties();
		parseAbstractGeometry(landUse.getLod1MultiSurface(), lod1Polygons, landColor);
		parseAbstractGeometry(landUse.getLod2MultiSurface(), lod2Polygons, landColor);
		parseAbstractGeometry(landUse.getLod3MultiSurface(), lod3Polygons, landColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, landColor);
	}

	@Override
	public void visit(AuxiliaryTrafficArea ata) {
		super.visit(ata);
		DeprecatedPropertiesOfAbstractThematicSurface deprecatedProperties = ata.getDeprecatedProperties();
		parseAbstractGeometry(ata.getLod2MultiSurface(), lod2Polygons, transportationColor);
		parseAbstractGeometry(ata.getLod3MultiSurface(), lod3Polygons, transportationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, transportationColor);
	}

	@Override
	public void visit(TrafficArea ta) {
		super.visit(ta);
		DeprecatedPropertiesOfAbstractThematicSurface deprecatedProperties = ta.getDeprecatedProperties();
		parseAbstractGeometry(ta.getLod2MultiSurface(), lod2Polygons, transportationColor);
		parseAbstractGeometry(ta.getLod3MultiSurface(), lod3Polygons, transportationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, transportationColor);
	}

	@Override
	public void visit(TransportationComplex tc) {
		super.visit(tc);
		DeprecatedPropertiesOfAbstractTransportationSpace deprecatedProperties = tc.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSurface(), lod1Polygons, transportationColor);
		parseAbstractGeometry(tc.getLod2MultiSurface(), lod2Polygons, transportationColor);
		parseAbstractGeometry(tc.getLod3MultiSurface(), lod3Polygons, transportationColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, transportationColor);
	}

	private Color getBuildingSurfaceColor(AbstractThematicSurface abs) {
		if (abs instanceof RoofSurface) {
			return roofColor;
		}
		if (abs instanceof GroundSurface) {
			return groundColor;
		}
		return wallColor;
	}

	@Override
	public void visit(AbstractFillingElement ao) {
		super.visit(ao);
		Color color = getOpeningColor(ao);
		parseAbstractGeometry(ao.getLod1Solid(), lod1Polygons, color);
		parseAbstractGeometry(ao.getLod2Solid(), lod2Polygons, color);
		parseAbstractGeometry(ao.getLod3Solid(), lod3Polygons, color);
		parseAbstractGeometry(ao.getLod2MultiSurface(), lod2Polygons, color);
		parseAbstractGeometry(ao.getLod3MultiSurface(), lod3Polygons, color);
		parseImplicitGeometry(ao.getLod1ImplicitRepresentation(), lod1Polygons, color);
		parseImplicitGeometry(ao.getLod2ImplicitRepresentation(), lod2Polygons, color);
		parseImplicitGeometry(ao.getLod3ImplicitRepresentation(), lod3Polygons, color);
	}

	private Color getOpeningColor(AbstractFillingElement ao) {
		if (ao instanceof Door) {
			return doorColor;
		}
		if (ao instanceof Window) {
			return windowColor;
		}
		return wallColor;
	}

	@Override
	public void visit(AbstractBridge ab) {
		super.visit(ab);
		DeprecatedPropertiesOfAbstractBridge deprecatedProperties = ab.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod1MultiSurface(), lod1Polygons, bridgeColor);
		parseAbstractGeometry(ab.getLod2MultiSurface(), lod2Polygons, bridgeColor);
		parseAbstractGeometry(ab.getLod3MultiSurface(), lod3Polygons, bridgeColor);
		parseAbstractGeometry(deprecatedProperties.getLod4MultiSurface(), lod4Polygons, bridgeColor);
		parseAbstractGeometry(ab.getLod1Solid(), lod1Polygons, bridgeColor);
		parseAbstractGeometry(ab.getLod2Solid(), lod2Polygons, bridgeColor);
		parseAbstractGeometry(ab.getLod3Solid(), lod3Polygons, bridgeColor);
		parseAbstractGeometry(deprecatedProperties.getLod4Solid(), lod4Polygons, bridgeColor);
	}
	
	@Override
	public void visit(BuildingInstallation bi) {
		super.visit(bi);
		DeprecatedPropertiesOfBuildingInstallation deprecatedProperties = bi.getDeprecatedProperties();
		parseAbstractGeometry(deprecatedProperties.getLod2Geometry(), lod2Polygons, wallColor);
		parseAbstractGeometry(deprecatedProperties.getLod3Geometry(), lod3Polygons, wallColor);
		parseAbstractGeometry(deprecatedProperties.getLod4Geometry(), lod4Polygons, wallColor);
		parseImplicitGeometry(bi.getLod2ImplicitRepresentation(), lod2Polygons, wallColor);
		parseImplicitGeometry(bi.getLod3ImplicitRepresentation(), lod3Polygons, wallColor);
		parseImplicitGeometry(deprecatedProperties.getLod4ImplicitRepresentation(), lod4Polygons, wallColor);
	}

	private void parseImplicitGeometry(ImplicitGeometryProperty irp, List<Polygon> polygons, Color color) {
		if (irp == null || irp.getObject() == null) {
			return;
		}
		ImplicitGeometry implicitGeometry = irp.getObject();
		parseAbstractGeometry(implicitGeometry.getRelativeGeometry(), polygons, color);
	}

	private void parseAbstractGeometry(GeometryProperty<? extends AbstractGeometry> geom, List<Polygon> polygons,
			Color color) {
		if (geom == null || geom.getObject() == null) {
			return;
		}
		currentColor = color;
		currentPolygons = polygons;
		geom.getObject().accept(this);
		currentColor = null;
		currentPolygons = null;
	}

	public List<Polygon> getLod1Polygons() {
		return lod1Polygons;
	}

	public List<Polygon> getLod2Polygons() {
		return lod2Polygons;
	}

	public List<Polygon> getLod3Polygons() {
		return lod3Polygons;
	}

	public List<Polygon> getLod4Polygons() {
		return lod4Polygons;
	}

	public void movePolygonsBy(Vector3d center) {
		movePolygonsBy(lod1Polygons, center);
		movePolygonsBy(lod2Polygons, center);
		movePolygonsBy(lod3Polygons, center);
		movePolygonsBy(lod4Polygons, center);
	}

	private void movePolygonsBy(List<Polygon> polygons, Vector3d center) {
		for (Polygon p : polygons) {
			for (Vector3d v : p.getExteriorRing().getVertices()) {
				movePointBy(v, center);
			}
			for (Ring r : p.getInteriorRings()) {
				for (Vector3d v : r.getVertices()) {
					movePointBy(v, center);
				}
			}
		}
	}

	private void movePointBy(Vector3d v, Vector3d center) {
		v.setX(v.getX() - center.getX());
		v.setY(v.getY() - center.getY());
		v.setZ(v.getZ() - center.getZ());
	}
}
